# import json
# import logging
# from langchain.callbacks.tracers.schemas import Run
# from singulr_client.clients import singulr_ingestion_client
# from singulr_client.mapper.env_generator import EnvironmentGenerator
# from singulr_client.mapper.span_generator import SpanGenerator
# from singulr_client.payload.ingestion_payload import IngestionData
# from singulr_client.trace.trace import TraceTree
# from singulr_client.utils import generate_trace_id
# from typing import Dict
#
#
# # Created by msinghal at 05/10/23
# class SingulrTracerGenerator():
#     def __init__(self, **kwargs):
#         super().__init__(**kwargs)
#         self.run_map: Dict[str, Run] = {}
#
#     def _log_trace_from_run(self, run: Run) -> None:
#         model_dict = SpanGenerator().process_model(run)
#         trace_id = generate_trace_id(model_dict)
#         environment = EnvironmentGenerator().process_environments(run)
#         root_span = SpanGenerator().process_span(run, trace_id)
#
#         if root_span is None:
#             return
#
#         model_trace = TraceTree(trace_id=trace_id,
#                                 environment=environment,
#                                 root_span=root_span)
#
#         trace = model_trace.to_json()
#         # json_string = json.dumps(trace)
#
#         ingestion_payload = IngestionData(trace_id, trace)
#         print("sending traces to ingestion service {}".format(json.dumps(ingestion_payload.to_dict())))
#         try:
#             response = singulr_ingestion_client.Ingestion_Client().ingest_payload(ingestion_payload.to_dict())
#         except:
#             import traceback
#             print("exception while tracing {}".format(traceback.print_exc()))
#             return None
#
#         if response.status_code == 200:
#             print("sent successfully")
#         else:
#             print(response.status_code)
#
#     def start_trace(self, run: Run) -> None:
#         """Start a trace for a run."""
#         if run.parent_run_id:
#             parent_run = self.run_map[str(run.parent_run_id)]
#             if parent_run:
#                 if hasattr(parent_run, "child_runs") and isinstance(parent_run.child_runs, list):
#                     parent_run.child_runs.append(run)
#                 else:
#                     setattr(parent_run, 'child_runs', [run])
#             else:
#                 logging.debug(f"Parent run with UUID {run.parent_run_id} not found.")
#         self.run_map[str(run.id)] = run
#
#     def end_trace(self, run: Run) -> None:
#         """End a trace for a run."""
#         if not run.parent_run_id:
#             self.run_map.get(str(run.id))
#             run.extra = self.run_map.get(str(run.id)).extra
#             self._persist_run(run)
#         else:
#             parent_run = self.run_map.get(str(run.parent_run_id))
#             if parent_run is None:
#                 logging.debug(f"Parent run with UUID {run.parent_run_id} not found.")
#             elif (
#                     hasattr(run, 'child_execution_order') and run.child_execution_order is not None
#                     and parent_run.child_execution_order is not None
#                     and run.child_execution_order > parent_run.child_execution_order
#             ):
#                 parent_run.child_execution_order = run.child_execution_order
#         self.run_map.pop(str(run.id))
#
#     def _persist_run(self, run: Run) -> None:
#         """Persist a run."""
#         self._log_trace_from_run(run)



# Created by msinghal at 07/09/23

from typing import Any, List, Optional, Union
from uuid import UUID
import json
from langchain.callbacks.tracers.langchain import LangChainTracer
from singulr_client.clients.langchain_run_client import LangChainRunClient
from singulr_client.trace.trace import TraceTree
from singulr_client.utils import generate_trace_id
from langchain.callbacks.tracers.schemas import Run
from singulr_client.clients import singulr_trace_client
from singulr_client.mapper.env_generator import EnvironmentGenerator
from singulr_client.mapper.span_generator import SpanGenerator
from singulr_client.payload.ingestion_payload import IngestionData
from singulr_client.clients.utils import get_runtime_environment

from typing import (
    Any,
    List,
    Optional,
    Union,
    cast,
)
def _get_client() -> LangChainRunClient:
    """Get the client."""
    return LangChainRunClient()


class SingulrTracer(LangChainTracer):
    def __init__(
            self,
            example_id: Optional[Union[UUID, str]] = None,
            project_name: Optional[str] = None,
            client: Optional[LangChainRunClient] = None,
            tags: Optional[List[str]] = None,
            use_threading: bool = True,
            **kwargs: Any,
    ) -> None:
        super().__init__(example_id=example_id,
                         project_name=project_name,
                         client=client or _get_client(),
                         tags=tags,
                         use_threading=use_threading,
                         **kwargs)

    def _log_trace_from_run(self, run: Run) -> None:
        model_dict = SpanGenerator().process_model(run)
        trace_id = generate_trace_id(model_dict)
        environment = EnvironmentGenerator().process_environments(run)
        root_span = SpanGenerator().process_span(run, trace_id)

        if root_span is None:
            return

        model_trace = TraceTree(trace_id=trace_id,
                                environment=environment,
                                root_span=root_span)

        trace = model_trace.to_json()
        # json_string = json.dumps(trace)

        ingestion_payload = IngestionData(trace_id, trace)
        print("sending traces to ingestion service {}".format(json.dumps(ingestion_payload.to_dict())))
        try:
            response = singulr_trace_client.Ingestion_Client().ingest_payload(ingestion_payload.to_dict())
        except:
            import traceback
            print("exception while tracing {}".format(traceback.print_exc()))
            return None

        if response.status_code == 200:
            print("sent successfully")
        else:
            print(response.status_code)

    def _persist_run(self, run: Run) -> None:
        """Persist a run."""
        runtime_env = get_runtime_environment()
        run.extra["runtime"] = runtime_env
        self._log_trace_from_run(run)



from singulr_client.span.primitive_attribute import BaseAttribute, AttributeType


class TokenUsage:
    def __init__(self):
        self.promptTokens: int = 0
        self.completionTokens: int = 0
        self.totalTokens: int = 0

    def get_prompt_tokens(self) -> int:
        return self.promptTokens

    def set_prompt_tokens(self, promptTokens: int):
        self.promptTokens = promptTokens

    def get_completion_tokens(self) -> int:
        return self.completionTokens

    def set_completion_tokens(self, completionTokens: int):
        self.completionTokens = completionTokens

    def get_total_tokens(self) -> int:
        return self.totalTokens

    def set_total_tokens(self, totalTokens: int):
        self.totalTokens = totalTokens


class LLMStats(BaseAttribute):
    LLM_OUTPUT = "llm_output"

    def __init__(self):
        self.subType: str = ""
        self.tokenUsage: TokenUsage = TokenUsage()
        self.modelName: str = ""

    def get_key(self) -> str:
        return self.LLM_OUTPUT

    def get_type(self):
        return AttributeType.LLM_STATS

    # Empty methods for serialization and deserialization
    def set_key(self, key: str):
        pass

    def set_type(self, type_):
        pass

    def get_sub_type(self) -> str:
        return self.subType

    def set_sub_type(self, subType: str):
        self.subType = subType

    def get_token_usage(self) -> TokenUsage:
        return self.tokenUsage

    def set_token_usage(self, tokenUsage: TokenUsage):
        self.tokenUsage = tokenUsage

    def get_model_name(self) -> str:
        return self.modelName

    def set_model_name(self, modelName: str):
        self.modelName = modelName



# Created by msinghal at 04/10/23
from __future__ import annotations

from datetime import datetime, timedelta
from enum import Enum
from typing import Any, Dict, List, Optional, Union
from uuid import UUID

from pydantic import BaseModel, Field, PrivateAttr, StrictBool, StrictFloat, StrictInt
from typing_extensions import Literal

SCORE_TYPE = Union[StrictBool, StrictInt, StrictFloat, None]
VALUE_TYPE = Union[Dict, StrictBool, StrictInt, StrictFloat, str, None]


class DataType(str, Enum):
    """Enum for dataset data types."""
    kv = "kv"
    llm = "llm"
    chat = "chat"


class DatasetBase(BaseModel):
    """Dataset base model."""

    name: str
    description: Optional[str] = None
    data_type: Optional[DataType] = None

    class Config:
        frozen = True


class DatasetCreate(DatasetBase):
    """Dataset create model."""

    id: Optional[UUID] = None
    created_at: datetime = Field(default_factory=datetime.utcnow)


class Dataset(DatasetBase):
    """Dataset ORM model."""

    id: UUID
    created_at: datetime
    modified_at: Optional[datetime] = Field(default=None)


class RunTypeEnum(str, Enum):
    """(Deprecated) Enum for run types. Use string directly."""

    tool = "tool"
    chain = "chain"
    llm = "llm"
    retriever = "retriever"
    embedding = "embedding"
    prompt = "prompt"
    parser = "parser"


class RunBase(BaseModel):
    """Base Run schema."""

    id: UUID
    name: str
    start_time: datetime
    run_type: str
    """The type of run, such as tool, chain, llm, retriever,
    embedding, prompt, parser."""
    end_time: Optional[datetime] = None
    extra: Optional[dict] = None
    error: Optional[str] = None
    serialized: Optional[dict]
    events: Optional[List[Dict]] = None
    inputs: dict
    outputs: Optional[dict] = None
    reference_example_id: Optional[UUID] = None
    parent_run_id: Optional[UUID] = None
    tags: Optional[List[str]] = None


class Run(RunBase):
    """Run schema when loading from the DB."""

    execution_order: int
    """The execution order of the run within a run trace."""
    session_id: Optional[UUID] = None
    """The project ID this run belongs to."""
    child_run_ids: Optional[List[UUID]] = None
    """The child run IDs of this run."""
    child_runs: Optional[List[Run]] = None
    """The child runs of this run, if instructed to load using the client
    These are not populated by default, as it is a heavier query to make."""
    feedback_stats: Optional[Dict[str, Any]] = None
    """Feedback stats for this run."""
    app_path: Optional[str] = None
    """Relative URL path of this run within the app."""
    _host_url: Optional[str] = PrivateAttr(default=None)

    def __init__(self, _host_url: Optional[str] = None, **kwargs: Any) -> None:
        """Initialize a Run object."""
        super().__init__(**kwargs)
        self._host_url = _host_url

    @property
    def url(self) -> Optional[str]:
        """URL of this run within the app."""
        if self._host_url and self.app_path:
            return f"{self._host_url}{self.app_path}"
        return None


class TracerSession(BaseModel):
    """TracerSession schema for the API.

    Sessions are also referred to as "Projects" in the UI.
    """

    id: UUID
    """The ID of the project."""
    start_time: datetime = Field(default_factory=datetime.utcnow)
    """The time the project was created."""
    name: Optional[str] = None
    """The name of the session."""
    extra: Optional[Dict[str, Any]] = None
    """Extra metadata for the project."""
    tenant_id: UUID
    """The tenant ID this project belongs to."""


class TracerSessionResult(TracerSession):
    """TracerSession schema returned when reading a project
    by ID. Sessions are also referred to as "Projects" in the UI."""

    run_count: Optional[int]
    """The number of runs in the project."""
    latency_p50: Optional[timedelta]
    """The median (50th percentile) latency for the project."""
    latency_p99: Optional[timedelta]
    """The 99th percentile latency for the project."""
    total_tokens: Optional[int]
    """The total number of tokens consumed in the project."""
    prompt_tokens: Optional[int]
    """The total number of prompt tokens consumed in the project."""
    completion_tokens: Optional[int]
    """The total number of completion tokens consumed in the project."""
    last_run_start_time: Optional[datetime]
    """The start time of the last run in the project."""
    feedback_stats: Optional[Dict[str, Any]]
    """Feedback stats for the project."""
    reference_dataset_ids: Optional[List[UUID]]
    """The reference dataset IDs this project's runs were generated on."""
    run_facets: Optional[List[Dict[str, Any]]]
    """Facets for the runs in the project."""



# Created by msinghal at 19/09/23
import time
class IngestionData:
    def __init__(self, id, payload):
        self.id = id
        self.source = "source-1"
        self.source_type = "LANG_CHAIN"
        self.timestamp = int(time.time() * 1000)
        self.type = "TRACE"
        self.serialization_type = "JSON"
        self.payload = payload


    def to_dict(self):
        return {
            "id": self.id,
            "source": self.source,
            "source_type": self.source_type,
            "timestamp": self.timestamp,
            "type": self.type,
            "serialization_type": self.serialization_type,
            "payload": self.payload
        }



# Created by msinghal at 04/10/23
# Created by msinghal at 23/09/23
import platform
import os, socket, psutil
from functools import lru_cache


def get_memory():
    try:
        memory_info = psutil.virtual_memory()
        total_memory = memory_info.total
    except ImportError:
        total_memory = os.sysconf('SC_PAGE_SIZE') * os.sysconf('SC_PHYS_PAGES')

    return total_memory


def get_process():
    current_pid = os.getpid()
    current_process = psutil.Process(current_pid)
    return {"process_id": current_pid,
            "process_name": current_process.name(),
            "process_status": current_process.status(),
            "process_start_time": current_process.create_time()
            }


@lru_cache(maxsize=1)
def get_runtime_environment() -> dict:
    """Get information about the environment."""
    # Lazy import to avoid circular imports
    from langchain import __version__
    current_process_info = get_process()
    return {
        "library_version": __version__,
        "langchain_version":__version__,
        "library": "langchain",
        "platform": platform.platform(),
        "runtime": "python",
        "runtime_version": platform.python_version(),
        "platform_architecure": platform.architecture(),
        "platform_node": platform.node(),
        "platform_version": platform.version(),
        "platform_cpu": os.cpu_count(),
        "platform_memory": get_memory(),
        "node_name": socket.gethostname(),
        "node_ip": socket.gethostbyname(socket.gethostname()),
        "node_fqdn": socket.getfqdn(),
        "process_name": current_process_info["process_name"],
        "process_id": current_process_info["process_id"],
        "process_status": current_process_info["process_status"],
        "process_start_time": current_process_info["process_start_time"],
    }



from pydantic import BaseModel, Field
import typing, hashlib, pickle
from langchain.callbacks.tracers.schemas import Run
from enum import Enum
import dataclasses
import json
from dataclasses import dataclass, field
from typing import TYPE_CHECKING, Any, Dict, List, Optional, Union
from singulr_client.utils import _safe_serialize
from singulr_client.env.environment import Environment


class StatusCode(str, Enum):
    SUCCESS = "SUCCESS"
    ERROR = "ERROR"

    def __str__(self) -> str:
        return str(self.value)


class StatusCode(str, Enum):
    SUCCESS = "SUCCESS"
    ERROR = "ERROR"

    def __str__(self) -> str:
        return str(self.value)


class SpanKind(str, Enum):
    LLM = "LLM"
    CHAIN = "CHAIN"
    AGENT = "AGENT"
    TOOL = "TOOL"

    def __str__(self) -> str:
        return str(self.value)


@dataclass()
class Result:
    inputs: Optional[Dict[str, Any]] = field(default=None)
    outputs: Optional[Dict[str, Any]] = field(default=None)


@dataclass()
class Span:
    span_id: Optional[str] = field(default=None)
    name: Optional[str] = field(default=None)
    parent_span_id: Optional[str] = field(default=None)
    trace_id: Optional[str] = field(default=None)
    start_time_millis: Optional[int] = field(default=None)
    end_time_millis: Optional[int] = field(default=None)
    status_code: Optional[StatusCode] = field(default=None)
    status_message: Optional[str] = field(default=None)
    type: Optional[str] = field(default=None)
    sub_type: Optional[str] = field(default=None)
    attributes: Optional[Dict[str, Any]] = field(default=None)
    results: Optional[List[Result]] = field(default=None)
    child_spans: Optional[List["Span"]] = field(default=None)

    def add_attribute(self, key: str, value: Any, at_type: str, type: str) -> None:
        # if self.attributes is None:
        #     self.attributes = {}
        # self.attributes[key] = value
        attribute = {
            "@type": at_type,
            "key": key,
            "type": value,
            "value": value
        }
        return attribute

    def add_named_result(self, inputs: Dict[str, Any], outputs: Dict[str, Any]) -> None:
        if self.results is None:
            self.results = []
        self.results.append(Result(inputs, outputs))

    def add_child_span(self, span: "Span") -> None:
        if self.child_spans is None:
            self.child_spans = []
        self.child_spans.append(span)


@dataclass()
class TraceTree:
    trace_id: str = field(default=None)
    environment: Environment = Field(default_factory={})
    root_span: Span = field(default=None)


class TraceAttribute:
    """Descriptor for accessing and setting attributes of the `Trace` class."""

    def __set_name__(self, owner: type, name: str) -> None:
        self.name = name

    def __get__(self, instance: "Trace", owner: type) -> Any:
        return getattr(instance._span, self.name)

    def __set__(self, instance: "Trace", value: Any) -> None:
        setattr(instance._span, self.name, value)


def _hash_id(s: str) -> str:
    return hashlib.md5(s.encode("utf-8")).hexdigest()[:16]


def generate_env_variables(run: Run):
    return {}


class SingulrTraceTree():
    """Media object for trace tree data.

    Arguments:
        root_span (Span): The root span of the trace tree.
        model_dict (dict, optional): A dictionary containing the model dump.
            NOTE: model_dict is a completely-user-defined dict. The UI will render
            a JSON viewer for this dict, giving special treatment to dictionaries
            with a `_kind` key. This is because model vendors have such different
            serialization formats that we need to be flexible here.
    """

    def __init__(
            self,
            root_span: Span,
            model_dict: typing.Optional[dict] = None,
            environment_info: Environment = None
    ):
        super().__init__()
        self._root_span = root_span
        self._model_dict = model_dict
        self._environment_info = environment_info

    def generate_trace_id(self, run: Run):
        json_str = _safe_serialize(run)
        return _hash_id(json_str)

    def to_json(self) -> dict:
        res = {}
        if self._model_dict is not None:
            res["trace_id"] = self.generate_trace_id(self._model_dict)
            res["model_dict_dumps"] = _safe_serialize(self._model_dict)
        res["root_span_dumps"] = _safe_serialize(dataclasses.asdict(self._root_span))
        environment = _safe_serialize(dataclasses.asdict(self._environment_info))
        trace_tree = {"trace_id": res["trace_id"], "root_span": res["root_span_dumps"], "environment": environment}
        print("trace_tree: \n {}".format(trace_tree))
        return trace_tree

    def is_bound(self) -> bool:
        return True



# Created by msinghal at 04/10/23
from __future__ import annotations
import singulr_client.clients.singulr_apis as singulr_apis
import logging
import os, json
import pandas as pd
import time
from datetime import datetime

from functools import lru_cache
from typing import (
    TYPE_CHECKING,
    Any,
    DefaultDict,
    Dict,
    Iterator,
    List,
    Mapping,
    Optional,
    Sequence,
    Tuple,
    Union,
    cast,
)

from singulr_client.clients.utils import get_runtime_environment

logger = logging.getLogger(__name__)

def _serialize_json(obj: Any) -> str:
    """Serialize an object to JSON.

    Parameters
    ----------
    obj : Any
        The object to serialize.

    Returns
    -------
    str
        The serialized JSON string.

    Raises
    ------
    TypeError
        If the object type is not serializable.
    """
    if isinstance(obj, datetime):
        return obj.isoformat()
    else:
        return str(obj)

class LangChainRunClient:
    """Client for interacting with the Singulr SDK"""
    def create_run(
            self,
            name: str,
            inputs: Dict[str, Any],
            run_type: str,
            *,
            execution_order: Optional[int] = None,
            **kwargs: Any,
    ) -> None:
        project_name = kwargs.pop(
            "project_name",
            kwargs.pop(
                "session_name",
                os.environ.get(
                    # TODO: Deprecate LANGCHAIN_SESSION
                    "LANGCHAIN_PROJECT",
                    os.environ.get("LANGCHAIN_SESSION", "default"),
                ),
            ),
        )
        run_create = {
            **kwargs,
            "session_name": project_name,
            "name": name,
            "inputs": inputs,
            "run_type": run_type,
            "execution_order": execution_order if execution_order is not None else 1,
        }
        run_extra = cast(dict, run_create.setdefault("extra", {}))
        runtime = run_extra.setdefault("runtime", {})

        runtime_env = get_runtime_environment()
        run_extra["runtime"] = {**runtime_env, **runtime}
        print("client is making call to local server to create run {}".format(run_create))
        file_name = str(run_create["id"]) + "-" + str(int(time.time()))
        df = pd.DataFrame.from_dict([run_create])
        df = df.astype(str)
        df.to_parquet("~/Downloads/traces/create_event/{}.parquet".format(file_name), engine='pyarrow')
        return {"message": "Run created successfully"}
        #singulr_apis.create_run(json.dumps(run_create, default=_serialize_json))

    def update_run(
            self,
            run_id,
            **kwargs: Any,
    ) -> None:
        print("client is making call to local server to update run {}".format(run_id))
        file_name = str(kwargs["id"]) + "-" + str(int(time.time()))
        df = pd.DataFrame.from_dict([kwargs])
        df = df.astype(str)
        df.to_parquet("~/Downloads/traces/update_event/{}.parquet".format(file_name), engine='pyarrow')
        #singulr_apis.update_run(run_id, json.dumps(kwargs, default=_serialize_json))
        return {"message": f"Run {run_id} updated successfully"}




# Created by msinghal at 12/09/23
from singulr_client.env.environment import MachineInfo
from singulr_client.env.environment import ProcessInfo, NetworkInfo
from singulr_client.env.environment import RuntimeEnvironment, Environment, RequestSecurityContext, SecurityContext
from langchain.callbacks.tracers.schemas import Run


class EnvironmentGenerator(object):

    def _extract_machine_info(self, run: Run):
        machine_info = {}
        machine_info["operating_system"] = run.extra["runtime"]["platform"]
        machine_info["os_version"] = run.extra["runtime"]["platform_version"]
        machine_info["total_memory"] = run.extra["runtime"]["platform_memory"]
        machine_info["cpu_info"] = run.extra["runtime"]["platform"]
        machine_info["num_cores"] = run.extra["runtime"]["platform_cpu"]
        machine_info["architecture"] = str(run.extra["runtime"]["platform_architecure"])
        return MachineInfo(**machine_info)

    def _extract_process_info(self, run: Run):
        process_info = {}
        process_info["process_path"] = run.extra["runtime"]["process_name"]
        process_info["process_id"] = run.extra["runtime"]['process_id']
        process_info["status"] = run.extra["runtime"]["process_status"]
        # process_info["process_start_time"] = run.extra["runtime"]["process_start_time"]
        return ProcessInfo(**process_info)

    # extract network information
    def _extract_network_info(self, run: Run):
        network_info = {}
        network_info["host_name"] = run.extra["runtime"]["node_name"]
        network_info["ip_address"] = run.extra["runtime"]["node_ip"]
        network_info["subnet"] = None
        return [NetworkInfo(**network_info)]

    def _extract_runtime_env(self, run: Run):
        runtime_info = {}
        runtime_info["lang_runtime"] = "python"
        runtime_info["lang_runtime_version"] = run.extra["runtime"]["langchain_version"]
        runtime_info["entry_point"] = "main.py"
        runtime_info["virtual_environment"] = "conda"
        runtime_info["environment"] = "llm-security"
        return RuntimeEnvironment(**runtime_info)

    def _extract_security_context(self, run: Run):
        security_info = {}
        security_info["user_id"] = "madan"
        security_info["group_id"] = "group-101"
        return SecurityContext(**security_info)


    def _extract_file_system_info(self, run: Run):
        file_system_info = {}
        file_system_info["working_directory"] = "/path/to/working/directory"
        file_system_info["total_disk_space_mbs"] = 102400
        file_system_info["file_system_paths"] = ["/data", "/var/logs"]

    def process_environments(self, run: Run):
        env = Environment(machine=self._extract_machine_info(run),
                          networks=self._extract_network_info(run),
                          process=self._extract_process_info(run),
                          file_system=self._extract_file_system_info(run),
                          security_context=self._extract_security_context(run),
                          request_security_context=RequestSecurityContext(),
                          runtime_environment=self._extract_runtime_env(run),
                          )
        return env



# Created by msinghal at 11/09/23
import abc
from enum import Enum
from typing import Type

from typing_extensions import TypedDict
import attr

from pydantic import BaseModel


class AttributeType(str, Enum):
    UNKNOWN = "UNKNOWN"
    PRIMITIVE_STRING = "PRIMITIVE_STRING"
    PRIMITIVE_INT = "PRIMITIVE_INT"
    LLM_INVOCATION = "LLM_INVOCATION"
    PROMPT_TEMPLATE = "PROMPT_TEMPLATE"
    LLM_STATS = "LLM_STATS"

    def __str__(self) -> str:
        return str(self.value)


class Attribute(metaclass=abc.ABCMeta):
    @abc.abstractmethod
    def get_key(self) -> str:
        pass

    @abc.abstractmethod
    def get_type(self) -> AttributeType:
        pass


class BaseAttribute(Attribute):
    def __init__(self, key: str, type: AttributeType):
        self.key = key
        self.type = type

    def get_key(self) -> str:
        pass

    def get_type(self) -> AttributeType:
        pass

    def to_dict(self):
        return {
            "key": self.key,
            "type": self.type.value,
        }


class PrimitiveAttribute(BaseAttribute, metaclass=abc.ABCMeta):
    def __init__(self, key: str, type: AttributeType) -> object:
        super().__init__(key, type)

    def get_key(self) -> str:
        return self.key

    def get_type(self) -> AttributeType:
        return self.type

    def to_dict(self):
        return {
            "key": self.key,
            "type": self.type.value,
        }


class StringAttribute(PrimitiveAttribute):
    def __init__(self, key: str, value: str) -> object:
        super().__init__(key, AttributeType.PRIMITIVE_STRING)
        self.value = value

    def to_json(self):
        return {
            "@type": "string",
            "key": self.key,
            "type": self.type.value,
            "value": self.value
        }

    def to_dict(self):
        return {
            "@type": "str",
            "key": self.key,
            "type": self.type.value,
            "value": self.value
        }


class IntAttribute(PrimitiveAttribute):
    def __init__(self, key: str, value: int) -> object:
        super().__init__(key, AttributeType.PRIMITIVE_INT)
        self.type = AttributeType.PRIMITIVE_INT
        self.value: int = value

    def to_json(self):
        return {
            "@type": "int",
            "key": self.key,
            "type": self.type.value,
            "value": self.value
        }

    def to_dict(self):
        return {
            "@type": "int",
            "key": self.key,
            "type": self.type.value,
            "value": self.value
        }


if __name__ == '__main__':
    s = IntAttribute(key="execution_order", value=1)
    debug = True



# Created by msinghal at 23/09/23
"""Generic utility functions."""
import platform, os
import subprocess
from functools import lru_cache
from typing import Any, Callable, Dict, List, Mapping, Optional, Tuple
import psutil
import socket

from requests import HTTPError, Response


class LangSmithAPIError(Exception):
    """An error occurred while communicating with the LangSmith API."""


class LangSmithUserError(Exception):
    """An error occurred while communicating with the LangSmith API."""


class LangSmithError(Exception):
    """An error occurred while communicating with the LangSmith API."""


class LangSmithConnectionError(Exception):
    """Couldn't connect to the LangSmith API."""


def xor_args(*arg_groups: Tuple[str, ...]) -> Callable:
    """Validate specified keyword args are mutually exclusive."""

    def decorator(func: Callable) -> Callable:
        def wrapper(*args: Any, **kwargs: Any) -> Callable:
            """Validate exactly one arg in each group is not None."""
            counts = [
                sum(1 for arg in arg_group if kwargs.get(arg) is not None)
                for arg_group in arg_groups
            ]
            invalid_groups = [i for i, count in enumerate(counts) if count != 1]
            if invalid_groups:
                invalid_group_names = [", ".join(arg_groups[i]) for i in invalid_groups]
                raise ValueError(
                    "Exactly one argument in each of the following"
                    " groups must be defined:"
                    f" {', '.join(invalid_group_names)}"
                )
            return func(*args, **kwargs)

        return wrapper

    return decorator


def raise_for_status_with_text(response: Response) -> None:
    """Raise an error with the response text."""
    try:
        response.raise_for_status()
    except HTTPError as e:
        raise ValueError(response.text) from e


def get_memory():
    try:
        memory_info = psutil.virtual_memory()
        total_memory = memory_info.total
    except ImportError:
        total_memory = os.sysconf('SC_PAGE_SIZE') * os.sysconf('SC_PHYS_PAGES')

    return total_memory

def get_process():
    current_pid = os.getpid()
    current_process = psutil.Process(current_pid)
    return {"process_id": current_pid,
            "process_name" : current_process.name(),
            "process_status": current_process.status(),
            "process_start_time" : current_process.create_time()
            }

@lru_cache
def get_runtime_environment() -> dict:
    """Get information about the environment."""
    # Lazy import to avoid circular imports
    from langsmith import __version__
    current_process_info = get_process()
    return {
        "sdk_version": __version__,
        "library": "langsmith",
        "platform": platform.platform(),
        "runtime": "python",
        "runtime_version": platform.python_version(),
        "langchain_version": get_langchain_environment(),
        "platform_architecure": platform.architecture(),
        "platform_node": platform.node(),
        "platform_version": platform.version(),
        "platform_cpu": os.cpu_count(),
        "platform_memory": get_memory(),
        "node_name": socket.gethostname(),
        "node_ip": socket.gethostbyname(socket.gethostname()),
        "node_fqdn": socket.getfqdn(),
        "process_name": current_process_info["process_name"],
        "process_id":current_process_info["process_id"],
        "process_status": current_process_info["process_status"],
        "process_start_time": current_process_info["process_start_time"],
    }


def _get_message_type(message: Mapping[str, Any]) -> str:
    if not message:
        raise ValueError("Message is empty.")
    if "lc" in message:
        if "id" not in message:
            raise ValueError(
                f"Unexpected format for serialized message: {message}"
                " Message does not have an id."
            )
        return message["id"][-1].replace("Message", "").lower()
    else:
        if "type" not in message:
            raise ValueError(
                f"Unexpected format for stored message: {message}"
                " Message does not have a type."
            )
        return message["type"]


def _get_message_fields(message: Mapping[str, Any]) -> Mapping[str, Any]:
    if not message:
        raise ValueError("Message is empty.")
    if "lc" in message:
        if "kwargs" not in message:
            raise ValueError(
                f"Unexpected format for serialized message: {message}"
                " Message does not have kwargs."
            )
        return message["kwargs"]
    else:
        if "data" not in message:
            raise ValueError(
                f"Unexpected format for stored message: {message}"
                " Message does not have data."
            )
        return message["data"]


def _convert_message(message: Mapping[str, Any]) -> Dict[str, Any]:
    """Extract message from a message object."""
    message_type = _get_message_type(message)
    message_data = _get_message_fields(message)
    return {"type": message_type, "data": message_data}


def get_messages_from_inputs(inputs: Mapping[str, Any]) -> List[Dict[str, Any]]:
    if "messages" in inputs:
        return [_convert_message(message) for message in inputs["messages"]]
    if "message" in inputs:
        return [_convert_message(inputs["message"])]
    raise ValueError(f"Could not find message(s) in run with inputs {inputs}.")


def get_message_generation_from_outputs(outputs: Mapping[str, Any]) -> Dict[str, Any]:
    if "generations" not in outputs:
        raise ValueError(f"No generations found in in run with output: {outputs}.")
    generations = outputs["generations"]
    if len(generations) != 1:
        raise ValueError(
            "Chat examples expect exactly one generation."
            f" Found {len(generations)} generations: {generations}."
        )
    first_generation = generations[0]
    if "message" not in first_generation:
        raise ValueError(
            f"Unexpected format for generation: {first_generation}."
            " Generation does not have a message."
        )
    return _convert_message(first_generation["message"])


def get_prompt_from_inputs(inputs: Mapping[str, Any]) -> str:
    if "prompt" in inputs:
        return inputs["prompt"]
    if "prompts" in inputs:
        prompts = inputs["prompts"]
        if len(prompts) == 1:
            return prompts[0]
        raise ValueError(
            f"Multiple prompts in run with inputs {inputs}."
            " Please create example manually."
        )
    raise ValueError(f"Could not find prompt in run with inputs {inputs}.")


def get_llm_generation_from_outputs(outputs: Mapping[str, Any]) -> str:
    if "generations" not in outputs:
        raise ValueError(f"No generations found in in run with output: {outputs}.")
    generations = outputs["generations"]
    if len(generations) != 1:
        raise ValueError(f"Multiple generations in run: {generations}")
    first_generation = generations[0]
    if "text" not in first_generation:
        raise ValueError(f"No text in generation: {first_generation}")
    return first_generation["text"]


@lru_cache
def get_docker_compose_command() -> List[str]:
    """Get the correct docker compose command for this system."""
    try:
        subprocess.check_call(
            ["docker", "compose", "--version"],
            stdout=subprocess.DEVNULL,
            stderr=subprocess.DEVNULL,
        )
        return ["docker", "compose"]
    except (subprocess.CalledProcessError, FileNotFoundError):
        try:
            subprocess.check_call(
                ["docker-compose", "--version"],
                stdout=subprocess.DEVNULL,
                stderr=subprocess.DEVNULL,
            )
            return ["docker-compose"]
        except (subprocess.CalledProcessError, FileNotFoundError):
            raise ValueError(
                "Neither 'docker compose' nor 'docker-compose'"
                " commands are available. Please install the Docker"
                " server following the instructions for your operating"
                " system at https://docs.docker.com/engine/install/"
            )


@lru_cache
def get_langchain_environment() -> Optional[str]:
    try:
        import langchain  # type: ignore

        return langchain.__version__
    except ImportError:
        return None


@lru_cache
def get_docker_version() -> Optional[str]:
    import subprocess

    try:
        docker_version = (
            subprocess.check_output(["docker", "--version"]).decode("utf-8").strip()
        )
    except FileNotFoundError:
        docker_version = "unknown"
    except:  # noqa
        return None
    return docker_version


@lru_cache
def get_docker_compose_version() -> Optional[str]:
    try:
        docker_compose_version = (
            subprocess.check_output(["docker-compose", "--version"])
            .decode("utf-8")
            .strip()
        )
    except FileNotFoundError:
        docker_compose_version = "unknown"
    except:  # noqa
        return None
    return docker_compose_version


@lru_cache
def _get_compose_command() -> Optional[List[str]]:
    try:
        compose_command = get_docker_compose_command()
    except ValueError as e:
        compose_command = [f"NOT INSTALLED: {e}"]
    except:  # noqa
        return None
    return compose_command


@lru_cache
def get_docker_environment() -> dict:
    """Get information about the environment."""
    compose_command = _get_compose_command()
    return {
        "docker_version": get_docker_version(),
        "docker_compose_command": " ".join(compose_command)
        if compose_command is not None
        else None,
        "docker_compose_version": get_docker_compose_version(),
    }



# Created by msinghal at 19/09/23
import requests


class Ingestion_Client(object):
    def __init__(self):
        self.injection_endpoint = "http://localhost:8080/api/ingestion/ingest"

    def ingest_payload(self, ingestion_payload):

        try:
            # Make an HTTP POST request to the Java REST endpoint with the data
            response = requests.post(self.injection_endpoint, json=ingestion_payload)  # Use json=data for sending JSON data

            # Check if the request was successful (status code 200)
            if response.status_code == 200:
                # Print the response content (JSON, XML, HTML, etc.)
                return response
            else:
                print(f"HTTP POST request failed with status code: {response.status_code}")
                return response

        except requests.exceptions.RequestException as e:
            print(f"An error occurred: {e}")
            return None


if __name__ == '__main__':
    ic = Ingestion_Client()
    import json
    # with open("simulated_sample.json", "r") as file:
    #     # Parse the JSON content
    #     data = json.load(file)
    # ic.ingest_payload(data)

    with open("b.json", "r") as file:
        # Parse the JSON content
        data = json.load(file)
    ic.ingest_payload(data)



from singulr_schemas import *
class Trace:
    """A simplification of WBTraceTree and Span to manage a trace - a collection of spans, their metadata and hierarchy.

    Args:
        name: (str) The name of the root span.
        kind: (str, optional) The kind of the root span.
        status_code: (str, optional) The status of the root span, either "error" or "success".
        status_message: (str, optional) Any status message associated with the root span.
        metadata: (dict, optional) Any additional metadata for the root span.
        start_time_ms: (int, optional) The start time of the root span in milliseconds.
        end_time_ms: (int, optional) The end time of the root span in milliseconds.
        inputs: (dict, optional) The named inputs of the root span.
        outputs: (dict, optional) The named outputs of the root span.
        model_dict: (dict, optional) A json serializable dictionary containing the model architecture details.

    Example:
        .. code-block:: python
        ```
        trace = Trace(
            name="My awesome Model",
            kind="LLM",
            status_code= "SUCCESS",
            metadata={"attr_1": 1, "attr_2": 2,},
            start_time_ms=int(round(time.time() * 1000)),
            end_time_ms=int(round(time.time() * 1000))+1000,
            inputs={"user": "How old is google?"},
            outputs={"assistant": "25 years old"},
            model_dict={"_kind": "openai", "api_type": "azure"}
              )
        run = wandb.init(project=<my_awesome_project>,)
        trace.log("my_trace")
        wandb.finish()
        ```
    """

    name = TraceAttribute()
    status_code = TraceAttribute()
    status_message = TraceAttribute()
    start_time_ms = TraceAttribute()
    end_time_ms = TraceAttribute()

    def __init__(
            self,
            name: str,
            kind: Optional[str] = None,
            status_code: Optional[str] = None,
            status_message: Optional[str] = None,
            metadata: Optional[dict] = None,
            start_time_ms: Optional[int] = None,
            end_time_ms: Optional[int] = None,
            inputs: Optional[dict] = None,
            outputs: Optional[dict] = None,
            model_dict: Optional[dict] = None,
    ):
        self._span = self._assert_and_create_span(
            name=name,
            kind=kind,
            status_code=status_code,
            status_message=status_message,
            metadata=metadata,
            start_time_ms=start_time_ms,
            end_time_ms=end_time_ms,
            inputs=inputs,
            outputs=outputs,
        )
        if model_dict is not None:
            assert isinstance(model_dict, dict), "Model dict must be a dictionary"
        self._model_dict = model_dict

    def _assert_and_create_span(
            self,
            name: str,
            kind: Optional[str] = None,
            status_code: Optional[str] = None,
            status_message: Optional[str] = None,
            metadata: Optional[dict] = None,
            start_time_ms: Optional[int] = None,
            end_time_ms: Optional[int] = None,
            inputs: Optional[dict] = None,
            outputs: Optional[dict] = None,
    ) -> Span:
        """Utility to assert the validity of the span parameters and create a span object.

        Args:
            name: The name of the span.
            kind: The kind of the span.
            status_code: The status code of the span.
            status_message: The status message of the span.
            metadata: Dictionary of metadata to be logged with the span.
            start_time_ms: Start time of the span in milliseconds.
            end_time_ms: End time of the span in milliseconds.
            inputs: Dictionary of inputs to be logged with the span.
            outputs: Dictionary of outputs to be logged with the span.

        Returns:
            A Span object.
        """
        if kind is not None:
            assert (
                    kind.upper() in SpanKind.__members__
            ), "Invalid span kind, can be one of 'LLM', 'AGENT', 'CHAIN', 'TOOL'"
            kind = SpanKind(kind.upper())
        if status_code is not None:
            assert (
                    status_code.upper() in StatusCode.__members__
            ), "Invalid status code, can be one of 'SUCCESS' or 'ERROR'"
            status_code = StatusCode(status_code.upper())
        if inputs is not None:
            assert isinstance(inputs, dict), "Inputs must be a dictionary"
        if outputs is not None:
            assert isinstance(outputs, dict), "Outputs must be a dictionary"
        if inputs or outputs:
            result = Result(inputs=inputs, outputs=outputs)
        else:
            result = None

        if metadata is not None:
            assert isinstance(metadata, dict), "Metadata must be a dictionary"

        return Span(
            name=name,
            span_kind=kind,
            status_code=status_code,
            status_message=status_message,
            attributes=metadata,
            start_time_ms=start_time_ms,
            end_time_ms=end_time_ms,
            results=[result] if result else None,
        )

    def add_child(
            self,
            child: "Trace",
    ) -> "Trace":
        """Utility to add a child span to the current span of the trace.

        Args:
            child: The child span to be added to the current span of the trace.

        Returns:
            The current trace object with the child span added to it.
        """
        self._span.add_child_span(child._span)
        if self._model_dict is not None and child._model_dict is not None:
            self._model_dict.update({child._span.name: child._model_dict})
        return self

    def add_inputs_and_outputs(self, inputs: dict, outputs: dict) -> "Trace":
        """Add a result to the span of the current trace.

        Args:
            inputs: Dictionary of inputs to be logged with the span.
            outputs: Dictionary of outputs to be logged with the span.

        Returns:
            The current trace object with the result added to it.
        """
        if self._span.results is None:
            result = Result(inputs=inputs, outputs=outputs)
            self._span.results = [result]
        else:
            result = Result(inputs=inputs, outputs=outputs)
            self._span.results.append(result)
        return self

    def add_metadata(self, metadata: dict) -> "Trace":
        """Add metadata to the span of the current trace."""
        if self._span.attributes is None:
            self._span.attributes = metadata
        else:
            self._span.attributes.update(metadata)
        return self

    @property
    def metadata(self) -> Optional[Dict[str, str]]:
        """Get the metadata of the trace.

        Returns:
            Dictionary of metadata.
        """
        return self._span.attributes

    @metadata.setter
    def metadata(self, value: Dict[str, str]) -> None:
        """Set the metadata of the trace.

        Args:
            value: Dictionary of metadata to be set.
        """
        if self._span.attributes is None:
            self._span.attributes = value
        else:
            self._span.attributes.update(value)

    @property
    def inputs(self) -> Optional[Dict[str, str]]:
        """Get the inputs of the trace.

        Returns:
            Dictionary of inputs.
        """
        return self._span.results[-1].inputs if self._span.results else None

    @inputs.setter
    def inputs(self, value: Dict[str, str]) -> None:
        """Set the inputs of the trace.

        Args:
            value: Dictionary of inputs to be set.
        """
        if self._span.results is None:
            result = Result(inputs=value, outputs={})
            self._span.results = [result]
        else:
            result = Result(inputs=value, outputs=self._span.results[-1].outputs)
            self._span.results.append(result)

    @property
    def outputs(self) -> Optional[Dict[str, str]]:
        """Get the outputs of the trace.

        Returns:
            Dictionary of outputs.
        """
        return self._span.results[-1].outputs if self._span.results else None

    @outputs.setter
    def outputs(self, value: Dict[str, str]) -> None:
        """Set the outputs of the trace.

        Args:
            value: Dictionary of outputs to be set.
        """
        if self._span.results is None:
            result = Result(inputs={}, outputs=value)
            self._span.results = [result]
        else:
            result = Result(inputs=self._span.results[-1].inputs, outputs=value)
            self._span.results.append(result)

    @property
    def kind(self) -> Optional[str]:
        """Get the kind of the trace.

        Returns:
            The kind of the trace.
        """
        return self._span.span_kind.value if self._span.span_kind else None

    @kind.setter
    def kind(self, value: str) -> None:
        """Set the kind of the trace.

        Args:
            value: The kind of the trace to be set.
        """
        assert (
                value.upper() in SpanKind.__members__
        ), "Invalid span kind, can be one of 'LLM', 'AGENT', 'CHAIN', 'TOOL'"
        self._span.span_kind = SpanKind(value.upper())


# Created by msinghal at 11/09/23
from typing import Optional, Dict
from dataclasses import field, dataclass
from singulr_client.env.environment import Environment
from singulr_client.span.span import Span
import json

class TraceTree:
    def __init__(self, trace_id: Optional[str] = None, environment: Optional[Environment] = None, root_span: Optional[Span] = None):
        self.trace_id = trace_id if trace_id is not None else self.trace_id
        self.environment = environment if environment is not None else self.environment
        self.root_span = root_span if root_span is not None else self.root_span

    def to_dict(self) -> Dict[str, any]:
        trace_tree_dict = {
            "trace_id": self.trace_id,
            "environment": self.environment.to_dict() if self.environment else None,
            "root_span": self.root_span.to_dict() if self.root_span else None
        }
        # Remove fields with None values
        trace_tree_dict = {k: v for k, v in trace_tree_dict.items() if v is not None}
        return trace_tree_dict

    def to_json(self) -> dict:
        trace_tree_dict =self.to_dict()
        json_string = json.dumps(trace_tree_dict)
        print("trace_tree dict: \n {}".format(trace_tree_dict))
        print("trace_tree json: \n {}".format(json_string))
        return json_string


# Created by msinghal at 11/09/23
# from pydantic import BaseModel, Field
# import typing, hashlib, pickle
# from langchain.callbacks.tracers.schemas import Run
from enum import Enum
# import dataclasses
# import json
# from dataclasses import dataclass, field
# from typing import TYPE_CHECKING, Any, Dict, List, Optional, Union
# from singulr_client.utils import _safe_serialize


class StatusCode(str, Enum):
    SUCCESS = "SUCCESS"
    ERROR = "ERROR"

    def __str__(self) -> str:
        return str(self.value)


class MachineInfo:
    def __init__(self, operating_system="Linux", os_version="Ubuntu 20.04", total_memory=8192, cpu_info="Intel Core i7", num_cores=4, architecture="x86_64", network_interfaces=None):
        self.operating_system = operating_system
        self.os_version = os_version
        self.total_memory = total_memory
        self.cpu_info = cpu_info
        self.num_cores = num_cores
        self.architecture = architecture
        self.network_interfaces = network_interfaces if network_interfaces is not None else [NetworkInterface()]

    def to_dict(self):
        return {
            "operating_system": self.operating_system,
            "os_version": self.os_version,
            "total_memory": self.total_memory,
            "cpu_info": self.cpu_info,
            "num_cores": self.num_cores,
            "architecture": self.architecture,
            "network_interfaces": [network_interface.to_dict() for network_interface in self.network_interfaces]
        }

class NetworkInterface:
    def __init__(self, name="en0", mac="bc:d0:74:0e:e8:2c"):
        self.name = name
        self.mac = mac
        self.network = NetworkInfo()


    def to_dict(self):
        return {
            'name': self.name,
            'mac': self.mac,
            'network': self.network.to_dict() if self.network else None
        }
    def get_name(self):
        return self.name

    def set_name(self, name):
        self.name = name

    def get_mac(self):
        return self.mac

    def set_mac(self, mac):
        self.mac = mac

    def get_network(self):
        return self.network

    def set_network(self, network):
        self.network = network
class NetworkInfo:
    def __init__(self, host_name="fin01.example.com", ip_address="192.168.2.5", subnet="192.168.2.0/24"):
        self.host_name = host_name
        self.ip_address = ip_address
        self.subnet = subnet

    def to_dict(self):
        return {
            "host_name": self.host_name,
            "ip_address": self.ip_address,
            "subnet": self.subnet
        }


class EnvironmentContext:
    def __init__(self, environment_variables=None, command_line_args=None, log_files=None):
        self.environment_variables = environment_variables if environment_variables is not None else {}
        self.command_line_args = command_line_args if command_line_args is not None else []
        self.log_files = log_files if log_files is not None else []

    def to_dict(self):
        return {
            "environment_variables": self.environment_variables,
            "command_line_args": self.command_line_args,
            "log_files": self.log_files
        }


class ProcessInfo:
    def __init__(self, process_path="python chat_bot_app.py", process_id=10000, status="running", listen_ports=None, environment_context=None):
        self.process_path = process_path
        self.process_id = process_id
        self.status = status
        self.listen_ports = listen_ports if listen_ports is not None else []
        self.environment_context = environment_context if environment_context is not None else EnvironmentContext()

    def to_dict(self):
        return {
            "process_path": self.process_path,
            "process_id": self.process_id,
            "status": self.status,
            "listen_ports": self.listen_ports,
            "environment_context": self.environment_context.to_dict()
        }


from typing import List, Dict, Optional, Any


class RuntimeEnvironment:
    def __init__(self, lang_runtime="python", lang_runtime_version="3.11.4", entry_point="main.py", virtual_environment="CONDA", environment="llm-security"):
        self.lang_runtime = lang_runtime
        self.lang_runtime_version = lang_runtime_version
        self.entry_point = entry_point
        self.virtual_environment = virtual_environment
        self.environment = environment

    def to_dict(self):
        return {
            "lang_runtime": self.lang_runtime,
            "lang_runtime_version": self.lang_runtime_version,
            "entry_point": self.entry_point,
            "virtual_environment": self.virtual_environment,
            "environment": self.environment
        }


class SecurityContext:
    def __init__(self, user_id="root", group_id="wheel"):
        self.user_id = user_id
        self.group_id = group_id

    def to_dict(self):
        return {
            "user_id": self.user_id,
            "group_id": self.group_id
        }


class RequestSecurityContext:
    def __init__(self, user_id="amit", group_id="engineering"):
        self.user_id = user_id
        self.group_id = group_id

    def to_dict(self):
        return {
            "user_id": self.user_id,
            "group_id": self.group_id
        }


class FileSystemInfo:
    def __init__(self, working_directory="/apps/app1", total_disk_space_mbs=102400, file_system_paths=None):
        self.working_directory = working_directory
        self.total_disk_space_mbs = total_disk_space_mbs
        self.file_system_paths = file_system_paths if file_system_paths is not None else []

    def to_dict(self):
        return {
            "working_directory": self.working_directory,
            "total_disk_space_mbs": self.total_disk_space_mbs,
            "file_system_paths": self.file_system_paths
        }


class DeploymentInfo:
    def __init__(self, name="SampleDeployment", version="1.0.0", deployment_environment="Production", metadata=None):
        self.name = name
        self.version = version
        self.deployment_environment = deployment_environment
        self.metadata = metadata if metadata is not None else []

    def to_dict(self):
        return {
            "name": self.name,
            "version": self.version,
            "deployment_environment": self.deployment_environment,
            "metadata": self.metadata
        }


class FrameworkInfo:
    def __init__(self, used_frameworks=None, dependencies=None):
        self.used_frameworks = used_frameworks if used_frameworks is not None else []
        self.dependencies = dependencies if dependencies is not None else []

    def to_dict(self):
        return {
            "used_frameworks": self.used_frameworks,
            "dependencies": self.dependencies
        }


class AppFrameworkMetadata:
    def __init__(self, key="frameworkKey1", value="frameworkValue1", tags=None):
        self.key = key
        self.value = value
        self.tags = tags if tags is not None else []

    def to_dict(self):
        return {
            "key": self.key,
            "value": self.value,
            "tags": self.tags
        }


class ApplicationFramework:
    def __init__(self, app_framework="LANG_CHAIN", version="1.0", app_framework_metadata=None):
        self.app_framework = app_framework
        self.version = version
        self.app_framework_metadata = app_framework_metadata if app_framework_metadata is not None else []

    def to_dict(self):
        return {
            "app_framework": self.app_framework,
            "version": self.version,
            "app_framework_metadata": [metadata.to_dict() for metadata in self.app_framework_metadata]
        }


class Metadata:
    def __init__(self, key="k1", value="value1", tags=None):
        self.key = key
        self.value = value
        self.tags = tags if tags is not None else []

    def to_dict(self):
        return {
            "key": self.key,
            "value": self.value,
            "tags": self.tags
        }


class ApplicationMetadata:
    def __init__(self, name="app1", version="1.0", description="", owner="madan", metadata=None):
        self.name = name
        self.version = version
        self.description = description
        self.owner = owner
        self.metadata = metadata if metadata is not None else []

    def to_dict(self):
        return {
            "name": self.name,
            "version": self.version,
            "description": self.description,
            "owner": self.owner,
            "metadata": [metadata.to_dict() for metadata in self.metadata]
        }


class Environment:
    def __init__(self, machine=None, networks=None, process=None, runtime_environment=None,
                 security_context=None, request_security_context=None,
                 file_system=None, deployment=None, framework=None,
                 application_framework=None, application_metadata=None):
        self.machine = machine if machine else MachineInfo()
        self.networks = networks if networks else [NetworkInfo()]
        self.process = process if process else ProcessInfo()
        self.runtime_environment = runtime_environment if runtime_environment else RuntimeEnvironment()
        self.security_context = security_context if security_context else SecurityContext()
        self.request_security_context = request_security_context if request_security_context else RequestSecurityContext()
        self.file_system = file_system if file_system else FileSystemInfo()
        self.deployment = deployment if deployment else DeploymentInfo()
        self.framework = framework if framework else FrameworkInfo()
        self.application_framework = application_framework if application_framework else ApplicationFramework()
        self.application_metadata = application_metadata if application_metadata else ApplicationMetadata()

    def to_dict(self):
        return {
            "machine": self.machine.to_dict() if self.machine is not None else None,
            "networks": [network.to_dict() for network in self.networks],
            "process": self.process.to_dict() if self.process is not None else None,
            "runtime_environment": self.runtime_environment.to_dict() if self.runtime_environment is not None else None,
            "security_context": self.security_context.to_dict() if self.security_context is not None else None,
            "request_security_context": self.request_security_context.to_dict() if self.request_security_context else None,
            "file_system": self.file_system.to_dict() if self.file_system is not None else None,
            "deployment": self.deployment.to_dict() if self.deployment is not None else None,
            "framework": self.framework.to_dict() if self.framework is not None else None,
            "application_framework": self.application_framework.to_dict() if self.application_framework is not None else None,
            "application_metadata": self.application_metadata.to_dict() if self.application_metadata is not None else None
        }

    def add_attribute(self, key: str, value: Any) -> None:
        if self.attributes is None:
            self.attributes = {}
        self.attributes[key] = value



import hashlib
import time
from typing import Any
from langchain.callbacks.tracers.schemas import Run

def _hash_id(s: str) -> str:
    return hashlib.md5(s.encode("utf-8")).hexdigest()[:16]


def _serialize_io(run_inputs: dict) -> dict:
    from google.protobuf.json_format import MessageToJson
    from google.protobuf.message import Message

    serialized_inputs = {}
    for key, value in run_inputs.items():
        if isinstance(value, Message):
            serialized_inputs[key] = MessageToJson(value)
        elif key == "input_documents":
            serialized_inputs.update(
                {f"input_document_{i}": json.dumps(doc) for i, doc in enumerate(value)}
            )
        else:
            serialized_inputs[key] = value
    return serialized_inputs


def _fallback_serialize(obj: Any) -> str:
    try:
        return f"<<non-serializable: {type(obj).__qualname__}>>"
    except Exception:
        return "<<non-serializable>>"


def _safe_serialize(obj: dict) -> str:
    try:
        return json.dumps(
            obj,
            skipkeys=True,
            default=_fallback_serialize,
        )
    except Exception:
        return "{}"

import json

class CustomEncoder(json.JSONEncoder):
    def default(self, obj):
        if hasattr(obj, 'to_dict') and callable(obj.to_dict):
            return obj.to_dict()
        elif isinstance(obj, list):
            return [self.default(item) for item in obj]
        return super().default(obj)

def generate_trace_id(run: Run):
    json_str = _safe_serialize(run)
    timestamp = str(time.time())  # Add a timestamp (current time) to the input string
    input_str = json_str + timestamp
    return _hash_id(input_str)


# Created by msinghal at 11/09/23
from enum import Enum
from typing import List, Optional, Dict, Any
from singulr_client.span.result import Result
from singulr_client.span.primitive_attribute import PrimitiveAttribute
from dataclasses import dataclass, field
from pydantic import Field, BaseModel

class SpanType(Enum):
    UNKNOWN = "UNKNOWN"
    LLM = "LLM"
    CHAIN = "CHAIN"
    TOOL = "TOOL"
    AGENT = "AGENT"

    def __str__(self) -> str:
        return str(self.value)


class StatusCode(str, Enum):
    SUCCESS = "SUCCESS"
    ERROR = "ERROR"

    def __str__(self) -> str:
        return str(self.value)



class Span(object):
    def __init__(
        self,
        span_id: Optional[str] = None,
        name: Optional[str] = None,
        parent_span_id: Optional[str] = None,
        trace_id: Optional[str] = None,
        start_time_millis: Optional[int] = None,
        end_time_millis: Optional[int] = None,
        status_code: Optional[StatusCode] = None,
        status_message: Optional[str] = None,
        type: Optional[SpanType] = None,
        sub_type: Optional[str] = None,
        attributes: Optional[List[PrimitiveAttribute]] = None,
        results: Optional[List[Result]] = None,
        child_spans: Optional[List["Span"]] = None,
    ):
        self.span_id = span_id
        self.name = name
        self.parent_span_id = parent_span_id
        self.trace_id = trace_id
        self.start_time_millis = start_time_millis
        self.end_time_millis = end_time_millis
        self.status_code = status_code
        self.status_message = status_message
        self.type = type
        self.sub_type = sub_type
        self.attributes = attributes if attributes is not None else []
        self.results = results if results is not None else []
        self.child_spans = child_spans if child_spans is not None else []

    def add_attribute(self, key: str, value: Any, at_type: str, type: str) -> None:
        # if self.attributes is None:
        #     self.attributes = {}
        # self.attributes[key] = value
        attr = {
            "@type": at_type,
            "key": key,
            "type": type,
            "value": value
        }
        return attr

    def add_named_result(self, inputs: Dict[str, Any], outputs: Dict[str, Any]) -> None:
        if self.results is None:
            self.results = []
        self.results.append(Result(inputs, outputs))

    def add_child_span(self, span: "Span") -> None:
        if self.child_spans is None:
            self.child_spans = []
        self.child_spans.append(span)

    def get_span_id(self) -> Optional[str]:
        return self.span_id

    def set_span_id(self, span_id: str) -> None:
        self.span_id = span_id

    def get_name(self) -> Optional[str]:
        return self.name

    def set_name(self, name: str) -> None:
        self.name = name

    def get_parent_span_id(self) -> Optional[str]:
        return self.parent_span_id

    def set_parent_span_id(self, parent_span_id: str) -> None:
        self.parent_span_id = parent_span_id

    def get_trace_id(self) -> Optional[str]:
        return self.trace_id

    def set_trace_id(self, trace_id: str) -> None:
        self.trace_id = trace_id

    def get_start_time_millis(self) -> Optional[int]:
        return self.start_time_millis

    def set_start_time_millis(self, start_time_millis: int) -> None:
        self.start_time_millis = start_time_millis

    def get_end_time_millis(self) -> Optional[int]:
        return self.end_time_millis

    def set_end_time_millis(self, end_time_millis: int) -> None:
        self.end_time_millis = end_time_millis

    def get_status_code(self) -> Optional[StatusCode]:
        return self.status_code

    def set_status_code(self, status_code: StatusCode) -> None:
        self.status_code = status_code

    def get_status_message(self) -> Optional[str]:
        return self.status_message

    def set_status_message(self, status_message: str) -> None:
        self.status_message = status_message

    def get_type(self) -> Optional[SpanType]:
        return self.type

    def set_type(self, type_: SpanType) -> None:
        self.type = type_

    def get_sub_type(self) -> Optional[str]:
        return self.sub_type

    def set_sub_type(self, sub_type: str) -> None:
        self.sub_type = sub_type

    def get_attributes(self) -> List[PrimitiveAttribute]:
        return self.attributes

    def set_attributes(self, attributes: List[PrimitiveAttribute]) -> None:
        self.attributes = attributes

    def get_results(self) -> List[Result]:
        return self.results

    def set_results(self, results: List[Result]) -> None:
        self.results = results

    def get_child_spans(self) -> List["Span"]:
        return self.child_spans

    def set_child_spans(self, child_spans: List["Span"]) -> None:
        self.child_spans = child_spans

    def to_dict(self):
        span_dict = {
            "span_id": self.span_id,
            "name": self.name,
            "parent_span_id": self.parent_span_id,
            "trace_id": self.trace_id,
            "start_time_millis": self.start_time_millis,
            "end_time_millis": self.end_time_millis,
            "status_code": self.status_code.value,
            "status_message": self.status_message,
            "type": self.type.value,
            "sub_type": self.sub_type,
            "attributes": [attr.to_dict() for attr in self.attributes],
            "results": [result.to_dict() for result in self.results],
            "child_spans": [span.to_dict() for span in self.child_spans],
        }
        # Remove fields with None values
        span_dict = {k: v for k, v in span_dict.items() if v is not None}
        return span_dict




# Created by msinghal at 12/09/23
import json
from langchain.callbacks.tracers.schemas import Run
from typing import Optional, Tuple, Dict, Any, List
from singulr_client.utils import _serialize_io
from singulr_client.span.result import Result
from singulr_client.span.content import TextDocument, GenerationDocument, DocumentMetadata, GenerationInfo, ContentType
from singulr_client.span.primitive_attribute import IntAttribute, StringAttribute
from singulr_client.span.span import SpanType, Span, StatusCode


class SpanGenerator(object):
    """Handles the conversion of a LangChain Runs."""

    def process_span(self, run: Run, trace_id: str) -> Optional["Span"]:
        """Converts a LangChain Run into a W&B Trace Span.
        :param run: The LangChain Run to convert.
        :return: The converted W&B Trace Span.
        """
        try:
            span = self._convert_lc_run_to_sg_span(run, trace_id)
            return span
        except Exception as e:
            import traceback
            print(traceback.format_exc())
            # TODO: handle exception and app centric error code generation
            return None

    def _convert_run_to_sg_span(self, run: Run, trace_id: str) -> "Span":
        """Base utility to create a span from a run.
        :param run: The run to convert.
        :return: The converted Span.
        """
        # attributes = {**run.extra} if run.extra else {}
        # attributes["execution_order"] = run.execution_order
        attribute = IntAttribute(key="execution_order", value=run.execution_order)
        return Span(
            span_id=str(run.id) if run.id is not None else None,
            name=run.name,
            start_time_millis=int(run.start_time.timestamp() * 1000),
            end_time_millis=int(run.end_time.timestamp() * 1000),
            status_code=StatusCode.SUCCESS
            if run.error is None
            else StatusCode.ERROR,
            status_message=run.error,
            attributes=[attribute],
            trace_id=trace_id
        )

    def _convert_llm_run_to_sg_span(self, run: Run, trace_id) -> "Span":
        """Converts a LangChain LLM Run into a W&B Trace Span.
        :param run: The LangChain LLM Run to convert.
        :return: The converted W&B Trace Span.
        """
        base_span = self._convert_run_to_sg_span(run, trace_id)

        # TODO: we will generic way to add attributes
        # if base_span.attributes is None:
        #     base_span.attributes = {}
        # base_span.attributes["llm_output"] = run.outputs.get("llm_output", {})

        for ndx, message in enumerate(run.inputs['messages'][0] or []):
            result = Result()
            content = message["kwargs"]['content']
            result.inputs = [TextDocument("prompt", content)]
            if (run.outputs is not None and len(run.outputs["generations"]) > ndx
                and len(run.outputs["generations"][ndx]) > 0):
                for g_i, gen in enumerate(run.outputs["generations"][ndx]):
                    gen_doc = GenerationDocument(f"gen_{g_i}", gen["text"])
                    if "generation_info" in gen:
                        gen_doc.generation_info = GenerationInfo(gen["generation_info"]['finish_reason'] if 'finish_reason' in gen else None)
                    else:
                        gen_doc.generation_info = GenerationInfo()
                    result.outputs.append(GenerationDocument(f"gen_{g_i}", gen["text"]))
            base_span.results.append(result)


        base_span.type = SpanType.LLM
        base_span.sub_type = run.name

        return base_span

    # if "input_documents" in run_inputs:
    #     docs = run_inputs["input_documents"]
    #     return {f"input_document_{i}": doc.json() for i, doc in enumerate(docs)}
    def _convert_chain_results_to_sg_results(self, run):
        result = Result()
        for key, input_data in run.inputs.items():
            if key == "input_documents":
                for i, doc in enumerate(input_data):
                    text_document = TextDocument(f"input_document_{i}", doc.page_content)
                    document_metadata = DocumentMetadata()
                    document_metadata.source = doc.metadata["source"]
                    document_metadata.pointer = doc.metadata["page"]
                    text_document.metadata = document_metadata
                    result.inputs.append(text_document)
            else:
                text_document = TextDocument(key, input_data)
                result.inputs.append(text_document)

        for key, output in run.outputs.items():
            text_document = TextDocument(key, output)
            result.outputs.append(text_document)
        return [result]

    def _convert_chain_run_to_sg_span(self, run: Run, trace_id) -> "Span":
        """Converts a LangChain Chain Run into a W&B Trace Span.
        :param run: The LangChain Chain Run to convert.
        :return: The converted W&B Trace Span.
        """
        base_span = self._convert_run_to_sg_span(run, trace_id)
        base_span.results = self._convert_chain_results_to_sg_results(run)
        base_span.child_spans = [
            self._convert_lc_run_to_sg_span(child_run, trace_id) for child_run in run.child_runs
        ]
        base_span.type = (
            SpanType.AGENT
            if "agent" in run.name.lower()
            else SpanType.CHAIN
        )
        base_span.sub_type = run.name

        return base_span

    def _convert_tool_run_to_sg_span(self, run: Run, trace_id) -> "Span":
        """Converts a LangChain Tool Run into a W&B Trace Span.
        :param run: The LangChain Tool Run to convert.
        :return: The converted W&B Trace Span.
        """
        base_span = self._convert_run_to_sg_span(run, trace_id)
        base_span.results = self._convert_chain_results_to_sg_results(run)
        base_span.child_spans = [
            self._convert_lc_run_to_sg_span(child_run) for child_run in run.child_runs
        ]
        base_span.span_kind = SpanType.TOOL

        return base_span

    def _convert_lc_run_to_sg_span(self, run: Run, trace_id: str) -> "Span":
        """Utility to convert any generic LangChain Run into a W&B Trace Span.
        :param run: The LangChain Run to convert.
        :return: The converted W&B Trace Span.
        """
        if run.run_type == "llm":
            return self._convert_llm_run_to_sg_span(run, trace_id)
        elif run.run_type == "chain":
            return self._convert_chain_run_to_sg_span(run, trace_id)
        elif run.run_type == "tool":
            return self._convert_tool_run_to_sg_span(run, trace_id)
        else:
            return self._convert_run_to_sg_span(run, trace_id)

    def process_model(self, run: Run) -> Optional[Dict[str, Any]]:
        """Utility to process a run for wandb model_dict serialization.
        :param run: The run to process.
        :return: The convert model_dict to pass to WBTraceTree.
        """
        try:
            data = json.loads(run.json())
            processed = self.flatten_run(data)
            keep_keys = (
                "id",
                "name",
                "serialized",
                "inputs",
                "outputs",
                "parent_run_id",
                "execution_order",
            )
            processed = self.truncate_run_iterative(processed, keep_keys=keep_keys)
            exact_keys, partial_keys = ("lc", "type"), ("api_key",)
            processed = self.modify_serialized_iterative(
                processed, exact_keys=exact_keys, partial_keys=partial_keys
            )
            output = self.build_tree(processed)
            return output
        except Exception as e:
            # TODO: handle exception
            return None

    def flatten_run(self, run: Dict[str, Any]) -> List[Dict[str, Any]]:
        """Utility to flatten a nest run object into a list of runs.
        :param run: The base run to flatten.
        :return: The flattened list of runs.
        """

        def flatten(child_runs: List[Dict[str, Any]]) -> List[Dict[str, Any]]:
            """Utility to recursively flatten a list of child runs in a run.
            :param child_runs: The list of child runs to flatten.
            :return: The flattened list of runs.
            """
            if child_runs is None:
                return []

            result = []
            for item in child_runs:
                child_runs = item.pop("child_runs", [])
                result.append(item)
                result.extend(flatten(child_runs))

            return result

        return flatten([run])

    def truncate_run_iterative(
            self, runs: List[Dict[str, Any]], keep_keys: Tuple[str, ...] = ()
    ) -> List[Dict[str, Any]]:
        """Utility to truncate a list of runs dictionaries to only keep the specified
            keys in each run.
        :param runs: The list of runs to truncate.
        :param keep_keys: The keys to keep in each run.
        :return: The truncated list of runs.
        """

        def truncate_single(run: Dict[str, Any]) -> Dict[str, Any]:
            """Utility to truncate a single run dictionary to only keep the specified
                keys.
            :param run: The run dictionary to truncate.
            :return: The truncated run dictionary
            """
            new_dict = {}
            for key in run:
                if key in keep_keys:
                    new_dict[key] = run.get(key)
            return new_dict

        return list(map(truncate_single, runs))

    def modify_serialized_iterative(
            self,
            runs: List[Dict[str, Any]],
            exact_keys: Tuple[str, ...] = (),
            partial_keys: Tuple[str, ...] = (),
    ) -> List[Dict[str, Any]]:
        """Utility to modify the serialized field of a list of runs dictionaries.
        removes any keys that match the exact_keys and any keys that contain any of the
        partial_keys.
        recursively moves the dictionaries under the kwargs key to the top level.
        changes the "id" field to a string "_kind" field that tells WBTraceTree how to
        visualize the run. promotes the "serialized" field to the top level.

        :param runs: The list of runs to modify.
        :param exact_keys: A tuple of keys to remove from the serialized field.
        :param partial_keys: A tuple of partial keys to remove from the serialized
            field.
        :return: The modified list of runs.
        """

        def remove_exact_and_partial_keys(obj: Dict[str, Any]) -> Dict[str, Any]:
            """Recursively removes exact and partial keys from a dictionary.
            :param obj: The dictionary to remove keys from.
            :return: The modified dictionary.
            """
            if isinstance(obj, dict):
                obj = {
                    k: v
                    for k, v in obj.items()
                    if k not in exact_keys
                       and not any(partial in k for partial in partial_keys)
                }
                for k, v in obj.items():
                    obj[k] = remove_exact_and_partial_keys(v)
            elif isinstance(obj, list):
                obj = [remove_exact_and_partial_keys(x) for x in obj]
            return obj

        def handle_id_and_kwargs(
                obj: Dict[str, Any], root: bool = False
        ) -> Dict[str, Any]:
            """Recursively handles the id and kwargs fields of a dictionary.
            changes the id field to a string "_kind" field that tells WBTraceTree how
            to visualize the run. recursively moves the dictionaries under the kwargs
            key to the top level.
            :param obj: a run dictionary with id and kwargs fields.
            :param root: whether this is the root dictionary or the serialized
                dictionary.
            :return: The modified dictionary.
            """
            if isinstance(obj, dict):
                if ("id" in obj or "name" in obj) and not root:
                    _kind = obj.get("id")
                    if not _kind:
                        _kind = [obj.get("name")]
                    obj["_kind"] = _kind[-1]
                    obj.pop("id", None)
                    obj.pop("name", None)
                    if "kwargs" in obj:
                        kwargs = obj.pop("kwargs")
                        for k, v in kwargs.items():
                            obj[k] = v
                for k, v in obj.items():
                    obj[k] = handle_id_and_kwargs(v)
            elif isinstance(obj, list):
                obj = [handle_id_and_kwargs(x) for x in obj]
            return obj

        def transform_serialized(serialized: Dict[str, Any]) -> Dict[str, Any]:
            """Transforms the serialized field of a run dictionary to be compatible
                with WBTraceTree.
            :param serialized: The serialized field of a run dictionary.
            :return: The transformed serialized field.
            """
            serialized = handle_id_and_kwargs(serialized, root=True)
            serialized = remove_exact_and_partial_keys(serialized)
            return serialized

        def transform_run(run: Dict[str, Any]) -> Dict[str, Any]:
            """Transforms a run dictionary to be compatible with WBTraceTree.
            :param run: The run dictionary to transform.
            :return: The transformed run dictionary.
            """
            transformed_dict = transform_serialized(run)

            serialized = transformed_dict.pop("serialized")
            for k, v in serialized.items():
                transformed_dict[k] = v

            _kind = transformed_dict.get("_kind", None)
            name = transformed_dict.pop("name", None)
            exec_ord = transformed_dict.pop("execution_order", None)

            if not name:
                name = _kind

            output_dict = {
                f"{exec_ord}_{name}": transformed_dict,
            }
            return output_dict

        return list(map(transform_run, runs))

    def build_tree(self, runs: List[Dict[str, Any]]) -> Dict[str, Any]:
        """Builds a nested dictionary from a list of runs.
        :param runs: The list of runs to build the tree from.
        :return: The nested dictionary representing the langchain Run in a tree
            structure compatible with WBTraceTree.
        """
        id_to_data = {}
        child_to_parent = {}

        for entity in runs:
            for key, data in entity.items():
                id_val = data.pop("id", None)
                parent_run_id = data.pop("parent_run_id", None)
                id_to_data[id_val] = {key: data}
                if parent_run_id:
                    child_to_parent[id_val] = parent_run_id

        for child_id, parent_id in child_to_parent.items():
            parent_dict = id_to_data[parent_id]
            parent_dict[next(iter(parent_dict))][
                next(iter(id_to_data[child_id]))
            ] = id_to_data[child_id][next(iter(id_to_data[child_id]))]

        root_dict = next(
            data for id_val, data in id_to_data.items() if id_val not in child_to_parent
        )

        return root_dict



# Created by msinghal at 12/09/23
from singulr_client.span.primitive_attribute import BaseAttribute, AttributeType
from typing import List
class Prompt(BaseAttribute):
    PROMPT = "prompt"

    def __init__(self):
        self.subType: str = ""
        self.template: str = ""
        self.inputVariables: List[str] = []
        self.templateFormat: str = ""
        self.type: str = self.get_type()
    def get_key(self) -> str:
        return self.PROMPT

    def get_type(self):
        return AttributeType.PROMPT_TEMPLATE

    # Empty methods for serialization and deserialization
    def set_key(self, key: str):
        pass

    def set_type(self, type_):
        pass

    def get_sub_type(self) -> str:
        return self.subType

    def set_sub_type(self, subType: str):
        self.subType = subType

    def get_template(self) -> str:
        return self.template

    def set_template(self, template: str):
        self.template = template

    def get_input_variables(self) -> List[str]:
        return self.inputVariables

    def set_input_variables(self, inputVariables: List[str]):
        self.inputVariables = inputVariables

    def get_template_format(self) -> str:
        return self.templateFormat

    def set_template_format(self, templateFormat: str):
        self.templateFormat = templateFormat


# from fastapi import FastAPI, Path
# from pydantic import BaseModel
# from singulr_client.singulrtracergenerator import SingulrTracerGenerator
# from typing import Dict, Any, Optional
# from run import Run
# import time
# import uvicorn
# import pandas as pd
# import json
#
# st = SingulrTracerGenerator()
#
#
# class RunCreate(BaseModel):
#     name: str
#     inputs: Dict[str, Any]
#     run_type: str
#     execution_order: Optional[int] = None
#     session_name: str
#
#
# class RunUpdate(BaseModel):
#     # Define the fields you want to update here.
#     # For example, if you want to update 'name' and 'inputs':
#     name: Optional[str] = None
#     inputs: Optional[Dict[str, Any]] = None
#
#
#
# def create_run(run_data: Run):
#     # Your create_run logic here
#     # Access run_data.name, run_data.inputs, run_data.run_type, etc.
#     # Access project_name if provided as a query parameter
#     run_data = json.dumps(json.loads(run_data))
#
#     #print("creating run id entry in database: {}".format(Run.dict(run_data)))
#     file_name = str(run_data["id"]) + "-" + str(int(time.time()))
#     st.start_trace(run_data)
#     df = pd.DataFrame.from_dict([run_data])
#     df = df.astype(str)
#     df.to_parquet("~/Downloads/traces/create_event/{}.parquet".format(file_name), engine='pyarrow')
#     return {"message": "Run created successfully"}
#
#
# def update_run(run_id: str, run_data: Run):
#     # Your update_run logic here
#     # Access run_id and run_data to perform the update
#     run_data = json.loads(run_data)
#     print("updating run id entry in database{}".format(run_id))
#     # print("updating run data {}".format(Run.dict(run_data)))
#     file_name = str(run_data.id) + "-" + str(int(time.time()))
#     st.end_trace(run_data)
#     df = pd.DataFrame.from_dict([run_data.dict()])
#     df = df.astype(str)
#     df.to_parquet("~/Downloads/traces/update_event/{}.parquet".format(file_name), engine='pyarrow')
#     return {"message": f"Run {run_id} updated successfully"}
#
#
# def get_run(run_id: str = Path(..., description="The ID of the run")):
#     # Your get_run logic here
#     # Access run_id to retrieve the run
#     print("acçessing run_id {}".format(run_id))
#     return {"message": f"Run {run_id} retrieved successfully"}
#
#
#
# if __name__ == '__main__':
#     uvicorn.run(app, host="0.0.0.0", port=8001)



from singulr_client.span.content import KeyedContent
from typing import TYPE_CHECKING, Any, Dict, List, Optional, Union


class Result:
    def __init__(self):
        self.inputs: List[KeyedContent] = []
        self.outputs: List[KeyedContent] = []

    def get_inputs(self) -> List[KeyedContent]:
        return self.inputs

    def set_inputs(self, inputs: List[KeyedContent]) -> None:
        self.inputs = inputs

    def get_outputs(self) -> List[KeyedContent]:
        return self.outputs

    def set_outputs(self, outputs: List[KeyedContent]) -> None:
        self.outputs = outputs

    def to_dict(self) -> Dict[str, any]:
        result_dict = {
            "inputs": [input_item.to_dict() for input_item in self.inputs],
            "outputs": [output_item.to_dict() for output_item in self.outputs]
        }
        return result_dict



# Created by msinghal at 11/09/23
import typing, dataclasses
from singulr_client.span.span import Span
from singulr_client.env.environment import Environment
from singulr_client.utils import _safe_serialize, _hash_id
from langchain.callbacks.tracers.schemas import Run
from singulr_client.trace.trace import TraceTree
from dataclasses import dataclass, field



class SingulrTraceTree():
    """Media object for trace tree data.

    Arguments:
        root_span (Span): The root span of the trace tree.
        model_dict (dict, optional): A dictionary containing the model dump.
            NOTE: model_dict is a completely-user-defined dict. The UI will render
            a JSON viewer for this dict, giving special treatment to dictionaries
            with a `_kind` key. This is because model vendors have such different
            serialization formats that we need to be flexible here.
    """

    def __init__(
            self,
            root_span: Span,
            model_dict: typing.Optional[dict] = None,
            environment_info: Environment = None,
            trace_id:str=None
    ):
        super().__init__()
        self._root_span = root_span
        self._model_dict = model_dict
        self._environment_info = environment_info
        self._trace_id = trace_id

    def to_json(self) -> dict:

        span_dict = self._root_span.to_dict()
        env_dict = self._environment_info.to_dict()
        environment = _safe_serialize(dataclasses.asdict(self._environment_info))
        trace_tree = {"trace_id":  self._trace_id, "root_span": res["root_span_dumps"], "environment": environment}
        print("trace_tree: \n {}".format(trace_tree))
        return trace_tree

    def is_bound(self) -> bool:
        return True


